#define PERL_NO_GET_CONTEXT
#define PERLIO_NOT_STDIO 0    /* For co-existence with stdio only */

#ifdef __cplusplus
extern "C" {
#endif
#include "EXTERN.h"
#include "perl.h"
#include "XSUB.h"
#ifdef __cplusplus
}
#endif

#include <string.h>
#include <stdio.h>

#define NEED_newRV_noinc
#include "ppport.h"

#define ISNOT_HASH_REF(roothash) ((roothash==NULL) || (!SvROK(roothash)) || (SvTYPE(SvRV(roothash)) != SVt_PVHV))

#define TEST_ARRAY_REF(path,msg) if ((path==NULL) || (!SvROK(path)) || (SvTYPE(SvRV(path)) != SVt_PVAV) ) {sv_dump(path); die(msg ": array pointer was expected but not found");}
#define TEST_NOTEMPTY_ARRAY_REF(path,pathlen,msg) if ((path==NULL) || (!SvROK(path)) || (SvTYPE(SvRV(path)) != SVt_PVAV) || ((pathlen = av_len((AV *)SvRV(path))) < 0) ) {sv_dump(path); die(msg ": notempty array pointer was expected but not found");}
#define TEST_HASH_REF(roothash) if (ISNOT_HASH_REF(roothash)) {die("first arg: hash pointer was expected but not found");}

static SV *
walk_path(pTHX_ SV* roothash, SV* path, I32 stepnum, int isCreate) {
   SV *noderef = roothash;
   I32 curstep;
   HV *node;
   SV **keyptr;
   HE *val;
   /*fprintf(stderr, "DEBUG: stepnum=%d\n",stepnum); */
   for(curstep = 0; curstep< stepnum; curstep++) {
     if (ISNOT_HASH_REF(noderef)) return &PL_sv_undef;
     node = (HV*) SvRV(noderef);
     keyptr = av_fetch((AV*) SvRV(path), curstep, 0);
     if (keyptr==NULL) {
       warn("Oops!:walk_path: fetching key from array failed!\n");
       return &PL_sv_undef;
     }
     val = hv_fetch_ent  (node, *keyptr, isCreate, 0);
     if (val==NULL) {
	 return &PL_sv_undef;
     }
     noderef = HeVAL(val);
     if ((noderef == NULL || !SvOK(noderef)) && isCreate) {
       val = hv_store_ent (node, *keyptr, newRV_noinc((SV*)newHV()), HeHASH(val));
       /*val = hv_store_ent (node, *keyptr, (SV*) newHV(), HeHASH(val)); */
       noderef = HeVAL(val);
     }
   }
   return noderef;
}

MODULE = Data::Array2ArrayMap::Hash::XSTree		PACKAGE = Data::Array2ArrayMap::Hash::XSTree

SV *
get_ref_if(roothash,path)
    SV *roothash;
    SV *path;
INIT:
    I32 pathlen;
    SV *leafref;
    TEST_HASH_REF(roothash);
    TEST_ARRAY_REF(path,"get_ref_if(path arg)");
CODE:
    pathlen = av_len((AV *)SvRV(path));
    leafref = walk_path(aTHX_ roothash, path, pathlen+1, 0);
    if (leafref==NULL || leafref == &PL_sv_undef) XSRETURN_UNDEF;
    RETVAL = SvREFCNT_inc(leafref);
OUTPUT:
    RETVAL

SV *
_node_at(roothash,path)
    SV *roothash;
    SV *path;
INIT:
    I32 pathlen;
    SV *leafref;
    TEST_HASH_REF(roothash);
    TEST_ARRAY_REF(path,"_node_at(path arg)");
CODE:
    pathlen = av_len((AV *)SvRV(path));
    leafref = walk_path(aTHX_ roothash, path, pathlen, 0);
    if (ISNOT_HASH_REF(leafref)) XSRETURN_UNDEF;
    RETVAL = SvREFCNT_inc(leafref);
OUTPUT:
    RETVAL

void
set(roothash,path,valarrayptr)
    SV *roothash;
    SV *path;
    SV *valarrayptr;
INIT:
    I32 pathlen;
    SV *leafref;
    HV *leafnode;
    SV **leafkeyptr;
    HE *leafval;
    AV *gotarray;
    AV *storearray;
    SV **gotptr;
    SV *got;
    I32 vallen;
    I32 i;
    TEST_HASH_REF(roothash);
    TEST_NOTEMPTY_ARRAY_REF(path,pathlen,"set:(path,)");
    TEST_ARRAY_REF(valarrayptr,"set:(,value): ");
CODE:
    leafref = walk_path(aTHX_ roothash, path, pathlen, 1);
    if ((leafref==NULL) || (!SvROK(leafref))) {
      sv_dump(leafref);
      die ("Oops!:set:internal error - can't store in non-reference node");
    }
    if (SvTYPE(SvRV(leafref)) != SVt_PVHV) {
      sv_dump(SvRV(leafref));
      die ("set:supplied path overrides previously stored data - store refused");
    }
    leafnode = (HV*) SvRV(leafref);
    leafkeyptr = av_fetch((AV*) SvRV(path), pathlen, 0);
    if (leafkeyptr==NULL) {
      die("Oops!:set: fetching key from array failed!\n");
      XSRETURN_UNDEF;
    }
    storearray = newAV();
    gotarray = (AV*) SvRV(valarrayptr);
    vallen = av_len(gotarray);
    av_extend(storearray, vallen);
    for (i=0; i<=vallen; i++) {
      gotptr = av_fetch(gotarray, i, 0);
      if (gotptr!=NULL) {
	got=*gotptr;
	SvREFCNT_inc(got);
	if (NULL==av_store(storearray, i, got)) SvREFCNT_dec(got);
      }
    }
    leafval = hv_fetch_ent (leafnode, *leafkeyptr, 1, 0);
leafval = hv_store_ent (leafnode, *leafkeyptr, newRV_noinc((SV*)storearray), HeHASH(leafval));
    if (leafval==NULL) {
      SvREFCNT_dec(HeVAL(leafval));
      die("Oops!:set: can't store the value!\n");
    }

    /* raw set; does not duplicate values in array */

void
rawput(roothash,path,storeval)
    SV *roothash;
    SV *path;
    SV *storeval;
INIT:
    I32 pathlen;
    SV *leafref;
    HV *leafnode;
    SV **leafkeyptr;
    HE *leafval;
    TEST_HASH_REF(roothash);
    TEST_NOTEMPTY_ARRAY_REF(path,pathlen,"set:(path,)");
CODE:
    leafref = walk_path(aTHX_ roothash, path, pathlen, 1);
    if (ISNOT_HASH_REF(leafref)) {
      sv_dump(leafref);
      die ("set:invalid path - can't store");
    }
    leafnode = (HV*) SvRV(leafref);
    leafkeyptr = av_fetch((AV*) SvRV(path), pathlen, 0);
    if (leafkeyptr==NULL) {
      die("Oops!:set: fetching key from array failed!\n");
      XSRETURN_UNDEF;
    }
    leafval = hv_fetch_ent (leafnode, *leafkeyptr, 1, 0);
    leafval = hv_store_ent (leafnode, *leafkeyptr, SvREFCNT_inc(storeval), HeHASH(leafval));
    if (leafval==NULL) {
      SvREFCNT_dec(HeVAL(leafval));
      die("Oops!:set: can't store the value!\n");
    }


SV*
get(roothash,path)
    SV *roothash;
    SV *path;
INIT:
    I32 pathlen;
    SV *leafref;
    I32 i;
    AV *leafarray;
    TEST_HASH_REF(roothash);
    TEST_NOTEMPTY_ARRAY_REF(path,pathlen,"get(path arg)");
PPCODE:
    leafref = walk_path(aTHX_ roothash, path, pathlen+1, 0);
    if (leafref==NULL || leafref == &PL_sv_undef) XSRETURN_EMPTY;
    if ((!SvROK(leafref)) || (SvTYPE(SvRV(leafref)) != SVt_PVAV)
       ) {
      warn("value: array pointer was expected but not found");
      XSRETURN_EMPTY;
    }
    pathlen = av_len((AV *)SvRV(leafref));
    switch (GIMME_V) {
      case G_VOID:
        XSRETURN_UNDEF
	;;
      case G_SCALAR:
	if (pathlen <0) {
	  XSRETURN_UNDEF;
	} else {
	  EXTEND(SP, 1);
	  PUSHs(sv_2mortal(newSVsv(*av_fetch((AV*) SvRV(leafref), 0, 0))));
	  /*XSRETURN(1); */
	}
	;;
      case G_ARRAY:
	if (pathlen>-1) {
	  EXTEND(SP, pathlen+1);
	  leafarray = (AV*) SvRV(leafref);
	  for (i=0; i<=pathlen; i++) {
	    PUSHs(sv_2mortal(newSVsv(*av_fetch(leafarray, i, 0))));
	  }
	  XSRETURN(pathlen+1);
	} else {
	  XSRETURN_EMPTY;
	}
	;;
    }

